/*
Copyright 2022 The KCP Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package options

import (
	"fmt"

	"github.com/spf13/pflag"

	kubeinformers "k8s.io/client-go/informers"
	"k8s.io/client-go/rest"

	kcpinformer "github.com/kcp-dev/kcp/pkg/client/informers/externalversions"
	apiexportoptions "github.com/kcp-dev/kcp/pkg/virtual/apiexport/options"
	"github.com/kcp-dev/kcp/pkg/virtual/framework"
	initializingworkspacesoptions "github.com/kcp-dev/kcp/pkg/virtual/initializingworkspaces/options"
	synceroptions "github.com/kcp-dev/kcp/pkg/virtual/syncer/options"
	workspacesoptions "github.com/kcp-dev/kcp/pkg/virtual/workspaces/options"
)

const virtualWorkspacesFlagPrefix = "virtual-workspaces-"

type Options struct {
	Workspaces             *workspacesoptions.Workspaces
	Syncer                 *synceroptions.Syncer
	APIExport              *apiexportoptions.APIExport
	InitializingWorkspaces *initializingworkspacesoptions.InitializingWorkspaces
}

func NewOptions() *Options {
	return &Options{
		Workspaces:             workspacesoptions.NewWorkspaces(),
		Syncer:                 synceroptions.NewSyncer(),
		APIExport:              apiexportoptions.NewAPIExport(),
		InitializingWorkspaces: initializingworkspacesoptions.New(),
	}
}

func (v *Options) Validate() []error {
	var errs []error

	errs = append(errs, v.Workspaces.Validate(virtualWorkspacesFlagPrefix)...)
	errs = append(errs, v.Syncer.Validate(virtualWorkspacesFlagPrefix)...)
	errs = append(errs, v.APIExport.Validate(virtualWorkspacesFlagPrefix)...)
	errs = append(errs, v.InitializingWorkspaces.Validate(virtualWorkspacesFlagPrefix)...)

	return errs
}

func (v *Options) AddFlags(fs *pflag.FlagSet) {
	v.Workspaces.AddFlags(fs, virtualWorkspacesFlagPrefix)
	v.InitializingWorkspaces.AddFlags(fs, virtualWorkspacesFlagPrefix)
}

func (o *Options) NewVirtualWorkspaces(
	config *rest.Config,
	rootPathPrefix string,
	wildcardKubeInformers kubeinformers.SharedInformerFactory,
	wildcardKcpInformers kcpinformer.SharedInformerFactory,
) (map[string]framework.VirtualWorkspace, error) {
	workspaces, err := o.Workspaces.NewVirtualWorkspaces(rootPathPrefix, config, wildcardKubeInformers, wildcardKcpInformers)
	if err != nil {
		return nil, err
	}

	syncer, err := o.Syncer.NewVirtualWorkspaces(rootPathPrefix, config, wildcardKcpInformers)
	if err != nil {
		return nil, err
	}

	apiexport, err := o.APIExport.NewVirtualWorkspaces(rootPathPrefix, config, wildcardKcpInformers)
	if err != nil {
		return nil, err
	}

	initializingworkspaces, err := o.InitializingWorkspaces.NewVirtualWorkspaces(rootPathPrefix, config, wildcardKcpInformers)
	if err != nil {
		return nil, err
	}

	all, err := merge(workspaces, syncer, apiexport, initializingworkspaces)
	if err != nil {
		return nil, err
	}
	return all, nil
}

func merge(sets ...map[string]framework.VirtualWorkspace) (map[string]framework.VirtualWorkspace, error) {
	workspaces := make(map[string]framework.VirtualWorkspace)
	for _, set := range sets {
		for name, w := range set {
			if _, ok := workspaces[name]; ok {
				return nil, fmt.Errorf("duplicate virtual workspace %q", name)
			}
			workspaces[name] = w
		}
	}
	return workspaces, nil
}
